/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jdbi.v3.core.mapper.reflect;

import java.lang.reflect.Field;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.OptionalInt;
import java.util.StringJoiner;
import java.util.concurrent.ConcurrentHashMap;

import org.jdbi.v3.core.annotation.Unmappable;
import org.jdbi.v3.core.mapper.ColumnMapper;
import org.jdbi.v3.core.mapper.Nested;
import org.jdbi.v3.core.mapper.PropagateNull;
import org.jdbi.v3.core.mapper.RowMapper;
import org.jdbi.v3.core.mapper.RowMapperFactory;
import org.jdbi.v3.core.mapper.SingleColumnMapper;
import org.jdbi.v3.core.mapper.reflect.internal.NullDelegatingMapper;
import org.jdbi.v3.core.qualifier.QualifiedType;
import org.jdbi.v3.core.qualifier.Qualifiers;
import org.jdbi.v3.core.statement.StatementContext;

import static java.lang.String.format;

import static org.jdbi.v3.core.mapper.ColumnMapper.getDefaultColumnMapper;
import static org.jdbi.v3.core.mapper.reflect.ReflectionMapperUtil.addPropertyNamePrefix;
import static org.jdbi.v3.core.mapper.reflect.ReflectionMapperUtil.anyColumnsStartWithPrefix;
import static org.jdbi.v3.core.mapper.reflect.ReflectionMapperUtil.findColumnIndex;
import static org.jdbi.v3.core.mapper.reflect.ReflectionMapperUtil.getColumnNames;

/**
 * A row mapper which maps the columns in a statement into an object, using reflection
 * to set fields on the object. All declared fields of the class and its superclasses
 * may be set. Nested properties are supported via the {@link Nested} annotation.
 *
 * The mapped class must have a default constructor.
 */
public final class FieldMapper<T> implements RowMapper<T> {
    private static final String DEFAULT_PREFIX = "";

    /**
     * Returns a mapper factory that maps to the given bean class
     *
     * @param type the mapped class
     * @return a mapper factory that maps to the given bean class
     */
    public static RowMapperFactory factory(Class<?> type) {
        return RowMapperFactory.of(type, FieldMapper.of(type));
    }

    /**
     * Returns a mapper factory that maps to the given bean class
     *
     * @param type the mapped class
     * @param prefix the column name prefix for each mapped field
     * @return a mapper factory that maps to the given bean class
     */
    public static RowMapperFactory factory(Class<?> type, String prefix) {
        return RowMapperFactory.of(type, FieldMapper.of(type, prefix));
    }

    /**
     * Returns a mapper for the given bean class
     *
     * @param <T> the type to map
     * @param type the mapped class
     * @return a mapper for the given bean class
     */
    public static <T> RowMapper<T> of(Class<T> type) {
        return FieldMapper.of(type, DEFAULT_PREFIX);
    }

    /**
     * Returns a mapper for the given bean class
     *
     * @param <T> the type to map
     * @param type the mapped class
     * @param prefix the column name prefix for each mapped field
     * @return a mapper for the given bean class
     */
    public static <T> RowMapper<T> of(Class<T> type, String prefix) {
        return new FieldMapper<>(type, prefix);
    }

    private final Class<T> type;
    private final String prefix;
    private final Map<Field, FieldMapper<?>> nestedMappers = new ConcurrentHashMap<>();

    private FieldMapper(Class<T> type, String prefix) {
        this.type = type;
        this.prefix = prefix;
    }

    @Override
    public T map(ResultSet rs, StatementContext ctx) throws SQLException {
        return specialize(rs, ctx).map(rs, ctx);
    }

    @Override
    public RowMapper<T> specialize(ResultSet rs, StatementContext ctx) throws SQLException {
        final List<String> columnNames = getColumnNames(rs);
        final List<ColumnNameMatcher> columnNameMatchers =
                ctx.getConfig(ReflectionMappers.class).getColumnNameMatchers();
        final List<String> unmatchedColumns = new ArrayList<>(columnNames);

        RowMapper<T> mapper = createSpecializedRowMapper(ctx, columnNames, columnNameMatchers, unmatchedColumns)
            .orElseThrow(() -> new IllegalArgumentException(format("Mapping fields for type %s didn't find any matching columns in result set", type)));

        if (ctx.getConfig(ReflectionMappers.class).isStrictMatching()
            && anyColumnsStartWithPrefix(unmatchedColumns, prefix, columnNameMatchers)) {
            throw new IllegalArgumentException(
                format("Mapping type %s could not match fields for columns: %s", type.getSimpleName(), unmatchedColumns));
        }

        return mapper;
    }

    private Optional<RowMapper<T>> createSpecializedRowMapper(StatementContext ctx,
                                               List<String> columnNames,
                                               List<ColumnNameMatcher> columnNameMatchers,
                                               List<String> unmatchedColumns) {
        final List<FieldData> fields = new ArrayList<>();

        for (Class<?> aType = type; aType != null; aType = aType.getSuperclass()) {
            for (Field field : aType.getDeclaredFields()) {
                Nested nested = field.getAnnotation(Nested.class);
                if (Optional.ofNullable(field.getAnnotation(Unmappable.class)).map(Unmappable::value).orElse(false)) {
                    continue;
                }

                if (nested == null) {
                    String paramName = addPropertyNamePrefix(prefix, paramName(field));

                    findColumnIndex(paramName, columnNames, columnNameMatchers, () -> debugName(field))
                        .ifPresent(index -> {
                            QualifiedType<?> fieldType = QualifiedType.of(field.getGenericType())
                                .withAnnotations(ctx.getConfig(Qualifiers.class).findFor(field));
                            ColumnMapper<?> mapper = ctx.findColumnMapperFor(fieldType)
                                .orElse(getDefaultColumnMapper());
                            fields.add(new FieldData(field, new SingleColumnMapper<>(mapper, index + 1)));
                            unmatchedColumns.remove(columnNames.get(index));
                        });
                } else {
                    String nestedPrefix = addPropertyNamePrefix(prefix, nested.value());

                    if (anyColumnsStartWithPrefix(columnNames, nestedPrefix, columnNameMatchers)) {
                        nestedMappers
                            .computeIfAbsent(field, f -> new FieldMapper<>(field.getType(), nestedPrefix))
                            .createSpecializedRowMapper(ctx, columnNames, columnNameMatchers, unmatchedColumns)
                            .ifPresent(mapper ->
                                fields.add(new FieldData(field, mapper)));
                    }
                }
            }
        }

        if (fields.isEmpty() && !columnNames.isEmpty()) {
            return Optional.empty();
        }

        fields.sort(Comparator.comparing(f -> f.propagateNull ? 1 : 0));

        RowMapper<T> boundMapper = new BoundFieldMapper(fields);
        OptionalInt propagateNullColumnIndex = locatePropagateNullColumnIndex(columnNames, columnNameMatchers);

        if (propagateNullColumnIndex.isPresent()) {
            return Optional.of(new NullDelegatingMapper<>(propagateNullColumnIndex.getAsInt() + 1, boundMapper));
        } else {
            return Optional.of(boundMapper);
        }
    }

    private OptionalInt locatePropagateNullColumnIndex(List<String> columnNames, List<ColumnNameMatcher> columnNameMatchers) {
        Optional<String> propagateNullColumn =
            Optional.ofNullable(type.getAnnotation(PropagateNull.class))
                .map(PropagateNull::value)
                .map(name -> addPropertyNamePrefix(prefix, name));

        if (!propagateNullColumn.isPresent()) {
            return OptionalInt.empty();
        }

        return findColumnIndex(propagateNullColumn.get(), columnNames, columnNameMatchers, propagateNullColumn::get);
    }

    private static String paramName(Field field) {
        return Optional.ofNullable(field.getAnnotation(ColumnName.class))
            .map(ColumnName::value)
            .orElseGet(field::getName);
    }

    private String debugName(Field field) {
        return format("%s.%s", type.getSimpleName(), field.getName());
    }

    public static boolean checkPropagateNullAnnotation(Field field) {
        final Optional<String> propagateNullValue = Optional.ofNullable(field.getAnnotation(PropagateNull.class)).map(PropagateNull::value);
        propagateNullValue.ifPresent(v -> {
            if (!v.isEmpty()) {
                throw new IllegalArgumentException(format("@PropagateNull does not support a value (%s) on a field (%s)", v, field.getName()));
            }
        });

        return propagateNullValue.isPresent();
    }

    private static class FieldData {

        FieldData(Field field, RowMapper<?> mapper) {
            this.field = field;
            this.mapper = mapper;
            propagateNull = checkPropagateNullAnnotation(field);
            isPrimitive = field.getType().isPrimitive();
        }

        final Field field;
        final RowMapper<?> mapper;
        final boolean propagateNull;
        final boolean isPrimitive;
    }

    class BoundFieldMapper implements RowMapper<T> {

        private final List<FieldData> fields;

        BoundFieldMapper(List<FieldData> fields) {
            this.fields = fields;
        }

        @Override
        public T map(ResultSet rs, StatementContext ctx) throws SQLException {
            T obj = construct();

            for (FieldData f : fields) {
                Object value = f.mapper.map(rs, ctx);
                if (f.propagateNull && (value == null || (f.isPrimitive && rs.wasNull()))) {
                    return null;
                }
                writeField(obj, f.field, value);
            }

            return obj;
        }

        private T construct() {
            try {
                return type.getDeclaredConstructor().newInstance();
            } catch (ReflectiveOperationException e) {
                throw new IllegalArgumentException(format("A type, %s, was mapped which was not instantiable", type.getName()), e);
            }
        }

        @SuppressWarnings("PMD.AvoidAccessibilityAlteration")
        private void writeField(T obj, Field field, Object value) {
            try {
                field.setAccessible(true);
                field.set(obj, value);
            } catch (IllegalAccessException e) {
                throw new IllegalArgumentException(format("Unable to access property, %s", field.getName()), e);
            }
        }

        @Override
        public String toString() {
            return new StringJoiner(", ", BoundFieldMapper.class.getSimpleName() + "[", "]")
                .add("type=" + type.getSimpleName())
                .add("prefix=" + prefix)
                .toString();
        }
    }
}

